@@ -4,6 +4,7 @@
 module Network.HTTP.Client.Core
     ( withResponse
     , httpLbs
+    , httpNoBody
     , httpRaw
     , responseOpen
     , responseClose
@@ -58,6 +59,13 @@ httpLbs req man = withResponse req man $ \res -> do
     bss <- brConsume $ responseBody res
     return res { responseBody = L.fromChunks bss }
 
+-- | A convenient wrapper around 'withResponse' which ignores the response
+-- body. This is useful, for example, when performing a HEAD request.
+--
+-- Since 0.3.2
+httpNoBody :: Request -> Manager -> IO (Response ())
+httpNoBody req man = withResponse req man $ return . void
+
 -- | Get a 'Response' without any redirect following.
 httpRaw
      :: Request
@@ -130,6 +138,11 @@ httpRaw req' m = do
 -- This function automatically performs any necessary redirects, as specified
 -- by the 'redirectCount' setting.
 --
+-- When implementing a (reverse) proxy using this function or relating
+-- functions, it's wise to remove Transfer-Encoding:, Content-Length:,
+-- Content-Encoding: and Accept-Encoding: from request and response
+-- headers to be relayed.
+--
 -- Since 0.1.0
 responseOpen :: Request -> Manager -> IO (Response BodyReader)
 responseOpen req0 manager = mWrapIOException manager $ do
@@ -180,29 +193,21 @@ httpRedirect
      -> IO (Response BodyReader)
 httpRedirect count0 http' req0 = go count0 req0 []
   where
-    go (-1) _ ress = throwIO . TooManyRedirects =<< mapM lbsResponse ress
+    go count _ ress | count < 0 = throwIO $ TooManyRedirects ress
     go count req' ress = do
         (res, mreq) <- http' req'
         case mreq of
             Just req -> do
-                {- FIXME
                 -- Allow the original connection to return to the
                 -- connection pool immediately by flushing the body.
                 -- If the response body is too large, don't flush, but
                 -- instead just close the connection.
                 let maxFlush = 1024
-                    readMay bs =
-                        case S8.readInt bs of
-                            Just (i, bs') | S.null bs' -> Just i
-                            _ -> Nothing
-                case lookup "content-length" (responseHeaders res) >>= readMay of
-                    Just i | i > maxFlush -> return ()
-                    _ -> void $ brReadSome (responseBody res) maxFlush
-                -}
+                lbs <- brReadSome (responseBody res) maxFlush
                 responseClose res
 
                 -- And now perform the actual redirect
-                go (count - 1) req (res:ress)
+                go (count - 1) req (res { responseBody = lbs }:ress)
             Nothing -> return res
 
 -- | Close any open resources associated with the given @Response@. In general,
