@@ -8,6 +8,7 @@ module Network.HTTP.Client.Manager
     ( ManagerSettings (..)
     , newManager
     , closeManager
+    , withManager
     , getConn
     , failedConnectionException
     , defaultManagerSettings
@@ -30,19 +31,21 @@ import Data.Text (Text)
 import qualified Data.Text as T
 
 import Control.Monad.IO.Class (MonadIO, liftIO)
-import Control.Exception (mask_, SomeException, catch, throwIO, fromException, mask, IOException, Exception (..), handle)
+import Control.Monad (unless)
+import Control.Exception (mask_, SomeException, bracket, catch, throwIO, fromException, mask, IOException, Exception (..), handle)
 import Control.Concurrent (forkIO, threadDelay)
 import Data.Time (UTCTime (..), Day (..), DiffTime, getCurrentTime, addUTCTime)
 import Control.DeepSeq (deepseq)
 
 import qualified Network.Socket as NS
 
-import Data.Default
 import Data.Maybe (mapMaybe)
 import System.IO (Handle)
 import System.Mem.Weak (Weak, deRefWeak)
+import Network.HTTP.Types (status200)
 import Network.HTTP.Client.Types
 import Network.HTTP.Client.Connection
+import Network.HTTP.Client.Headers (parseStatusHeaders)
 
 -- | Default value for @ManagerSettings@.
 --
@@ -52,7 +55,8 @@ defaultManagerSettings = ManagerSettings
     { managerConnCount = 10
     , managerRawConnection = return openSocketConnection
     , managerTlsConnection = return $ \_ _ _ -> throwIO TlsNotSupported
-    , managerResponseTimeout = Just 5000000
+    , managerTlsProxyConnection = return $ \_ _ _ _ _ -> throwIO TlsNotSupported
+    , managerResponseTimeout = Just 30000000
     , managerRetryableException = \e ->
         case fromException e of
             Just (_ :: IOException) -> True
@@ -64,6 +68,7 @@ defaultManagerSettings = ManagerSettings
                     -- we open a new connection under these circumstances, we
                     -- check for the NoResponseDataReceived exception.
                     Just NoResponseDataReceived -> True
+                    Just IncompleteHeaders -> True
                     _ -> False
     , managerWrapIOException =
         let wrapper se =
@@ -122,6 +127,7 @@ newManager :: ManagerSettings -> IO Manager
 newManager ms = do
     rawConnection <- managerRawConnection ms
     tlsConnection <- managerTlsConnection ms
+    tlsProxyConnection <- managerTlsProxyConnection ms
     mapRef <- I.newIORef (Just Map.empty)
     wmapRef <- I.mkWeakIORef mapRef $ closeManager' mapRef
     _ <- forkIO $ reap wmapRef
@@ -131,6 +137,7 @@ newManager ms = do
             , mResponseTimeout = managerResponseTimeout ms
             , mRawConnection = rawConnection
             , mTlsConnection = tlsConnection
+            , mTlsProxyConnection = tlsProxyConnection
             , mRetryableException = managerRetryableException ms
             , mWrapIOException = managerWrapIOException ms
             }
@@ -243,8 +250,14 @@ closeManager' connsRef = mask_ $ do
     m <- I.atomicModifyIORef connsRef $ \x -> (Nothing, x)
     mapM_ (nonEmptyMapM_ safeConnClose) $ maybe [] Map.elems m
 
+-- | Create, use and close a 'Manager'.
+--
+-- Since 0.2.1
+withManager :: ManagerSettings -> (Manager -> IO a) -> IO a
+withManager settings = bracket (newManager settings) closeManager
+
 safeConnClose :: Connection -> IO ()
-safeConnClose ci = connectionClose ci `catch` \(_::SomeException) -> return ()
+safeConnClose ci = connectionClose ci `catch` \(_ :: IOException) -> return ()
 
 nonEmptyMapM_ :: Monad m => (a -> m ()) -> NonEmptyList a -> m ()
 nonEmptyMapM_ f (One x _) = f x
@@ -277,11 +290,21 @@ getManagedConn man key open = mask $ \restore -> do
     -- reference to track what we want to do. By default, we say not to reuse
     -- it, that way if an exception is thrown, the connection won't be reused.
     toReuseRef <- I.newIORef DontReuse
+    wasReleasedRef <- I.newIORef False
 
     -- When the connection is explicitly released, we update our toReuseRef to
     -- indicate what action should be taken, and then call release.
-    let connRelease Reuse = putSocket man key ci
-        connRelease DontReuse = connectionClose ci
+    let connRelease r = do
+            I.writeIORef toReuseRef r
+            releaseHelper
+
+        releaseHelper = mask $ \restore -> do
+            wasReleased <- I.atomicModifyIORef wasReleasedRef $ \x -> (True, x)
+            unless wasReleased $ do
+                toReuse <- I.readIORef toReuseRef
+                restore $ case toReuse of
+                    Reuse -> putSocket man key ci
+                    DontReuse -> connectionClose ci
 
     return (connRelease, ci, isManaged)
 
@@ -302,18 +325,40 @@ getConnDest req =
 getConn :: Request
         -> Manager
         -> IO (ConnRelease, Connection, ManagedConn)
-getConn req m =
-    getManagedConn m (ConnKey connKeyHost connport (secure req)) $
-        go connaddr connhost connport
+getConn req m
+    -- Stop Mac OS X from getting high:
+    -- https://github.com/snoyberg/http-client/issues/40#issuecomment-39117909
+    | S8.null h = throwIO $ InvalidDestinationHost h
+    | otherwise =
+        getManagedConn m (ConnKey connKeyHost connport (secure req)) $
+            wrapConnectExc $ go connaddr connhost connport
   where
     h = host req
     (useProxy, connhost, connport) = getConnDest req
     (connaddr, connKeyHost) =
-        case (hostAddress req, useProxy{-, socksProxy req-}) of
-            (Just ha, False{-, Nothing-}) -> (Just ha, HostAddress ha)
+        case (hostAddress req, useProxy) of
+            (Just ha, False) -> (Just ha, HostAddress ha)
             _ -> (Nothing, HostName $ T.pack connhost)
+
+    wrapConnectExc = handle $ \e ->
+        throwIO $ FailedConnectionException2 connhost connport (secure req)
+            (toException (e :: IOException))
     go =
         case (secure req, useProxy) of
             (False, _) -> mRawConnection m
             (True, False) -> mTlsConnection m
-            -- FIXME (True, True) -> getSslProxyConn (checkCerts m h) (clientCertificates req) h (port req)
+            (True, True) ->
+                let ultHost = host req
+                    ultPort = port req
+                    connstr = S8.concat
+                        [ "CONNECT "
+                        , ultHost
+                        , ":"
+                        , S8.pack $ show ultPort
+                        , " HTTP/1.1\r\n\r\n"
+                        ]
+                    parse conn = do
+                        sh@(StatusHeaders status _ _) <- parseStatusHeaders conn
+                        unless (status == status200) $
+                            throwIO $ ProxyConnectException ultHost ultPort $ Left $ S8.pack $ show sh
+                 in mTlsProxyConnection m connstr parse
