@@ -1,8 +1,10 @@
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFoldable #-}
 {-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
 {-# LANGUAGE RankNTypes #-}
 module Network.HTTP.Client.Types
-    ( BodyReader (..)
+    ( BodyReader
     , Connection (..)
     , StatusHeaders (..)
     , HttpException (..)
@@ -34,9 +36,11 @@ import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
 import Blaze.ByteString.Builder (Builder, fromLazyByteString, fromByteString, toLazyByteString)
 import Data.Int (Int64)
-import Data.Default
+import Data.Default.Class
+import Data.Foldable (Foldable)
 import Data.Monoid
 import Data.Time (UTCTime)
+import Data.Traversable (Traversable)
 import qualified Data.List as DL
 import Network.Socket (HostAddress)
 import Data.IORef
@@ -44,33 +48,31 @@ import qualified Network.Socket as NS
 import qualified Data.IORef as I
 import qualified Data.Map as Map
 import Data.Text (Text)
+import Data.Streaming.Zlib (ZlibException)
 
--- | An abstraction for representing an incoming response body coming from the
--- server. Data provided by this abstraction has already been gunzipped and
+-- | An @IO@ action that represents an incoming response body coming from the
+-- server. Data provided by this action has already been gunzipped and
 -- de-chunked, and respects any content-length headers present.
 --
--- Since 0.1.0
-data BodyReader = BodyReader
-    { brRead :: !(IO S.ByteString)
-    -- ^ Get a single chunk of data from the response body, or an empty
-    -- bytestring if no more data is available.
-    --
-    -- Since 0.1.0
-    , brComplete :: !(IO Bool)
-    }
+-- The action gets a single chunk of data from the response body, or an empty
+-- bytestring if no more data is available.
+--
+-- Since 0.4.0
+type BodyReader = IO S.ByteString
 
 data Connection = Connection
-    { connectionRead :: !(IO S.ByteString)
+    { connectionRead :: IO S.ByteString
       -- ^ If no more data, return empty.
-    , connectionUnread :: !(S.ByteString -> IO ())
+    , connectionUnread :: S.ByteString -> IO ()
       -- ^ Return data to be read next time.
-    , connectionWrite :: !(S.ByteString -> IO ())
+    , connectionWrite :: S.ByteString -> IO ()
       -- ^ Send data to server
-    , connectionClose :: !(IO ())
+    , connectionClose :: IO ()
     }
+    deriving T.Typeable
 
-data StatusHeaders = StatusHeaders !Status !HttpVersion !RequestHeaders
-    deriving (Show, Eq, Ord)
+data StatusHeaders = StatusHeaders Status HttpVersion RequestHeaders
+    deriving (Show, Eq, Ord, T.Typeable)
 
 data HttpException = StatusCodeException Status ResponseHeaders CookieJar
                    | InvalidUrlException String String
@@ -82,6 +84,7 @@ data HttpException = StatusCodeException Status ResponseHeaders CookieJar
                    | OverlongHeaders
                    | ResponseTimeout
                    | FailedConnectionException String Int -- ^ host/port
+                   | FailedConnectionException2 String Int Bool SomeException -- ^ host/port/secure
                    | ExpectedBlankAfter100Continue
                    | InvalidStatusLine S.ByteString
                    | InvalidHeader S.ByteString
@@ -99,6 +102,11 @@ data HttpException = StatusCodeException Status ResponseHeaders CookieJar
                    --
                    -- Since 1.9.4
                    | IncompleteHeaders
+                   | InvalidDestinationHost S.ByteString
+                   | HttpZlibException ZlibException
+                   -- ^
+                   --
+                   -- Since 0.3
     deriving (Show, T.Typeable)
 instance Exception HttpException
 
@@ -117,10 +125,10 @@ data Cookie = Cookie
   , cookie_secure_only :: Bool
   , cookie_http_only :: Bool
   }
-  deriving (Read, Show)
+  deriving (Read, Show, T.Typeable)
 
 newtype CookieJar = CJ { expose :: [Cookie] }
-  deriving (Read, Show)
+  deriving (Read, Show, T.Typeable)
 
 -- This corresponds to step 11 of the algorithm described in Section 5.3 \"Storage Model\"
 instance Eq Cookie where
@@ -155,8 +163,8 @@ instance Monoid CookieJar where
 -- | Define a HTTP proxy, consisting of a hostname and port number.
 
 data Proxy = Proxy
-    { proxyHost :: !S.ByteString -- ^ The host name of the HTTP proxy.
-    , proxyPort :: !Int -- ^ The port number of the HTTP proxy.
+    { proxyHost :: S.ByteString -- ^ The host name of the HTTP proxy.
+    , proxyPort :: Int -- ^ The port number of the HTTP proxy.
     }
     deriving (Show, Read, Eq, Ord, T.Typeable)
 
@@ -170,11 +178,12 @@ data Proxy = Proxy
 --
 -- Since 0.1.0
 data RequestBody
-    = RequestBodyLBS !L.ByteString
-    | RequestBodyBS !S.ByteString
-    | RequestBodyBuilder !Int64 !Builder
-    | RequestBodyStream !Int64 !(GivesPopper ())
-    | RequestBodyStreamChunked !(GivesPopper ())
+    = RequestBodyLBS L.ByteString
+    | RequestBodyBS S.ByteString
+    | RequestBodyBuilder Int64 Builder
+    | RequestBodyStream Int64 (GivesPopper ())
+    | RequestBodyStreamChunked (GivesPopper ())
+    deriving T.Typeable
 instance Monoid RequestBody where
     mempty = RequestBodyBS S.empty
     mappend x0 y0 =
@@ -354,8 +363,9 @@ data Request = Request
     --
     -- Since 0.1.0
     , responseTimeout :: Maybe Int
-    -- ^ Number of microseconds to wait for a response. If @Nothing@, will wait
-    -- indefinitely. Default: 5 seconds.
+    -- ^ Number of microseconds to wait for a response. If
+    -- @Nothing@, will wait indefinitely. Default: use
+    -- 'managerResponseTimeout' (which by default is 5 seconds).
     --
     -- Since 0.1.0
     , getConnectionWrapper :: Maybe Int
@@ -379,8 +389,10 @@ data Request = Request
     --
     -- Since 0.1.0
     }
+    deriving T.Typeable
 
 data ConnReuse = Reuse | DontReuse
+    deriving T.Typeable
 
 type ConnRelease = ConnReuse -> IO ()
 
@@ -390,36 +402,36 @@ data ManagedConn = Fresh | Reused
 --
 -- Since 0.1.0
 data Response body = Response
-    { responseStatus :: !Status
+    { responseStatus :: Status
     -- ^ Status code of the response.
     --
     -- Since 0.1.0
-    , responseVersion :: !HttpVersion
+    , responseVersion :: HttpVersion
     -- ^ HTTP version used by the server.
     --
     -- Since 0.1.0
-    , responseHeaders :: !ResponseHeaders
+    , responseHeaders :: ResponseHeaders
     -- ^ Response headers sent by the server.
     --
     -- Since 0.1.0
-    , responseBody :: !body
+    , responseBody :: body
     -- ^ Response body sent by the server.
     --
     -- Since 0.1.0
-    , responseCookieJar :: !CookieJar
+    , responseCookieJar :: CookieJar
     -- ^ Cookies set on the client after interacting with the server. If
     -- cookies have been disabled by setting 'cookieJar' to @Nothing@, then
     -- this will always be empty.
     --
     -- Since 0.1.0
-    , responseClose' :: !ResponseClose
+    , responseClose' :: ResponseClose
     -- ^ Releases any resource held by this response. If the response body
     -- has not been fully read yet, doing so after this call will likely
     -- be impossible.
     --
     -- Since 0.1.0
     }
-    deriving (Show, Eq, T.Typeable, Functor)
+    deriving (Show, Eq, T.Typeable, Functor, Foldable, Traversable)
 
 newtype ResponseClose = ResponseClose { runResponseClose :: IO () }
     deriving T.Typeable
@@ -433,33 +445,38 @@ instance Eq ResponseClose where
 --
 -- Since 0.1.0
 data ManagerSettings = ManagerSettings
-    { managerConnCount :: !Int
+    { managerConnCount :: Int
       -- ^ Number of connections to a single host to keep alive. Default: 10.
       --
       -- Since 0.1.0
-    , managerRawConnection :: !(IO (Maybe NS.HostAddress -> String -> Int -> IO Connection))
+    , managerRawConnection :: IO (Maybe NS.HostAddress -> String -> Int -> IO Connection)
       -- ^ Create an insecure connection.
       --
       -- Since 0.1.0
-    , managerTlsConnection :: !(IO (Maybe NS.HostAddress -> String -> Int -> IO Connection))
+    -- FIXME in the future, combine managerTlsConnection and managerTlsProxyConnection
+    , managerTlsConnection :: IO (Maybe NS.HostAddress -> String -> Int -> IO Connection)
       -- ^ Create a TLS connection. Default behavior: throw an exception that TLS is not supported.
       --
       -- Since 0.1.0
-    , managerResponseTimeout :: !(Maybe Int)
+    , managerTlsProxyConnection :: IO (S.ByteString -> (Connection -> IO ()) -> Maybe NS.HostAddress -> String -> Int -> IO Connection)
+      -- ^ Create a TLS proxy connection. Default behavior: throw an exception that TLS is not supported.
+      --
+      -- Since 0.2.2
+    , managerResponseTimeout :: Maybe Int
       -- ^ Default timeout (in microseconds) to be applied to requests which do
       -- not provide a timeout value.
       --
       -- Default is 5 seconds
       --
       -- Since 0.1.0
-    , managerRetryableException :: !(SomeException -> Bool)
+    , managerRetryableException :: SomeException -> Bool
     -- ^ Exceptions for which we should retry our request if we were reusing an
     -- already open connection. In the case of IOExceptions, for example, we
     -- assume that the connection was closed on the server and therefore open a
     -- new one.
     --
     -- Since 0.1.0
-    , managerWrapIOException :: !(forall a. IO a -> IO a)
+    , managerWrapIOException :: forall a. IO a -> IO a
     -- ^ Action wrapped around all attempted @Request@s, usually used to wrap
     -- up exceptions in library-specific types.
     --
@@ -467,6 +484,7 @@ data ManagerSettings = ManagerSettings
     --
     -- Since 0.1.0
     }
+    deriving T.Typeable
 
 -- | Keeps track of open connections for keep-alive.
 --
@@ -474,29 +492,32 @@ data ManagerSettings = ManagerSettings
 --
 -- Since 0.1.0
 data Manager = Manager
-    { mConns :: !(I.IORef (Maybe (Map.Map ConnKey (NonEmptyList Connection))))
+    { mConns :: I.IORef (Maybe (Map.Map ConnKey (NonEmptyList Connection)))
     -- ^ @Nothing@ indicates that the manager is closed.
-    , mMaxConns :: !Int
+    , mMaxConns :: Int
     -- ^ This is a per-@ConnKey@ value.
-    , mResponseTimeout :: !(Maybe Int)
+    , mResponseTimeout :: Maybe Int
     -- ^ Copied from 'managerResponseTimeout'
-    , mRawConnection :: !(Maybe NS.HostAddress -> String -> Int -> IO Connection)
-    , mTlsConnection :: !(Maybe NS.HostAddress -> String -> Int -> IO Connection)
-    , mRetryableException :: !(SomeException -> Bool)
-    , mWrapIOException :: !(forall a. IO a -> IO a)
+    , mRawConnection :: Maybe NS.HostAddress -> String -> Int -> IO Connection
+    , mTlsConnection :: Maybe NS.HostAddress -> String -> Int -> IO Connection
+    , mTlsProxyConnection :: S.ByteString -> (Connection -> IO ()) -> Maybe NS.HostAddress -> String -> Int -> IO Connection
+    , mRetryableException :: SomeException -> Bool
+    , mWrapIOException :: forall a. IO a -> IO a
     }
+    deriving T.Typeable
 
 data NonEmptyList a =
-    One !a !UTCTime |
-    Cons !a !Int !UTCTime !(NonEmptyList a)
+    One a UTCTime |
+    Cons a Int UTCTime (NonEmptyList a)
+    deriving T.Typeable
 
 -- | Hostname or resolved host address.
 data ConnHost =
-    HostName !Text |
-    HostAddress !NS.HostAddress
-    deriving (Eq, Show, Ord)
+    HostName Text |
+    HostAddress NS.HostAddress
+    deriving (Eq, Show, Ord, T.Typeable)
 
 -- | @ConnKey@ consists of a hostname, a port and a @Bool@
 -- specifying whether to use SSL.
-data ConnKey = ConnKey !ConnHost !Int !Bool
-    deriving (Eq, Show, Ord)
+data ConnKey = ConnKey ConnHost Int Bool
+    deriving (Eq, Show, Ord, T.Typeable)
